# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Gonçalo Silva
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

from gettext import gettext as _

from gi.repository import Gtk, GLib, Handy, GObject

from apostrophe.settings import Settings
from apostrophe.preview_layout_switcher import PreviewLayout
from apostrophe.preview_window import PreviewWindow
from .tweener import Tweener

class PreviewRenderer(GObject.Object):
    """Renders the preview according to the user selected mode."""

    __gtype_name__ = "PreviewRenderer"

    preview_layout = GObject.Property(type=int, default=0)
    preview_window_title = GObject.Property(type=str, default="")

    def __init__(
            self, main_window, text_view, flap):
        super().__init__()
        self.main_window = main_window
        self.main_window.connect("delete-event", self.on_window_closed)
        self.main_window.connect("notify::title", self.on_window_title_changed)
        self.text_view = text_view

        self.window_size_cache = None

        self.settings = Settings.new()
        self.window = None
        self.preview_stack = self.main_window.preview_stack

        # we may get the preview layout changed underneath us, so we store
        # a cache for proper hidding the previous layout
        self.preview_layout_cache = None

        self.flap = flap

        self.connect("notify::preview-layout", self.update_mode)

    def show(self):
        """Show the preview, depending on the currently selected mode."""

        # Windowed preview: create a window and show the preview in it.
        if self.preview_layout == PreviewLayout.WINDOWED and not self.window:
            # Create transient window of the main window.
            self.window = PreviewWindow()
            self.window.connect("delete-event", self.on_window_closed)

            self.window.preview_box.pack_start(self.preview_stack, False, True, 0)

            self.bind_property("preview_window_title", self.window, "title")

            # Position it next to the main window.
            width, height = self.main_window.get_size()
            self.window.resize(width, height)
            x, y = self.main_window.get_position()
            if x is not None and y is not None:
                self.main_window.move(x, y)
                self.window.move(x + width + 16, y)

            self.window.show()

        elif self.preview_layout == PreviewLayout.HALF_WIDTH:
            self.window_size_cache = self.main_window.get_size()
            requested_width_tw = Tweener(self.flap,
                                        self.flap.set_size_request,
                                        self.text_view.get_min_width(),
                                        self.text_view.get_min_width() * 2 + 2,
                                        250, 
                                        setter_args = [-1])
            requested_width_tw.start()
            self.flap.set_reveal_flap(True)

        else:
            # if it's the first time we open the preview, set
            # the corresponding flap mode
            if self.preview_layout_cache is None:
                self.update_mode()
            self.flap.set_reveal_flap(True)

    def load_webview(self, webview):
        webview.show()
        self.main_window.preview_stack.add(webview)
        self.main_window.preview_stack.set_visible_child(webview)

    def hide(self):
        """Hide the preview, depending on the currently selected mode."""
        self.flap.set_reveal_flap(False)

        # Windowed preview: remove preview and destroy window.
        if self.preview_layout_cache == PreviewLayout.WINDOWED:
            self.main_window.present()
            self.window.preview_box.remove(self.preview_stack)
            self.window.destroy()
            self.window = None
        else:
            # Half-width/height previews: remove preview and reset size
            # requirements.
            if self.preview_layout_cache == PreviewLayout.HALF_WIDTH:
                self.shrink_window()

    def shrink_window(self):
        self.flap.set_size_request(self.text_view.get_min_width(), -1)
        resize_tw = Tweener(self.main_window,
                            self.main_window.resize,
                            self.main_window.get_size()[0],
                            self.window_size_cache[0],
                            250, 
                            setter_args = [self.window_size_cache[1]])
        resize_tw.start()

    def update_mode(self, *args, web_view=None):
        """Update preview mode
        """

        def set_flap_mode(*args, **kwargs):
            # TODO: use structural pattern matching in python3.10
            if c and self.flap.get_reveal_progress() != 0:
                return

            if c:
                self.main_window.flap.disconnect(c)

            if self.preview_layout == PreviewLayout.FULL_WIDTH:
                if self.preview_layout_cache == PreviewLayout.HALF_WIDTH:
                    self.shrink_window()
                self.flap.set_fold_policy(Handy.FlapFoldPolicy.ALWAYS)
                self.flap.set_orientation(Gtk.Orientation.HORIZONTAL)

            elif self.preview_layout == PreviewLayout.HALF_WIDTH:
                self.window_size_cache = self.main_window.get_size()
                requested_width_tw = Tweener(self.flap,
                                            self.flap.set_size_request,
                                            self.text_view.get_min_width(),
                                            self.text_view.get_min_width() * 2 + 2,
                                            250, 
                                            setter_args = [-1])
                requested_width_tw.start()

                self.flap.set_fold_policy(Handy.FlapFoldPolicy.NEVER)
                self.flap.set_orientation(Gtk.Orientation.HORIZONTAL)

            elif self.preview_layout == PreviewLayout.HALF_HEIGHT:
                self.flap.set_fold_policy(Handy.FlapFoldPolicy.NEVER)
                self.flap.set_orientation(Gtk.Orientation.VERTICAL)

            else:
                raise ValueError("Unknown preview mode {}".format(self.preview_layout))

            # don't automatically show the preview when syncing the preview-layout property first time
            if self.preview_layout_cache is not None:
                self.show()
            self.preview_layout_cache = self.preview_layout

        def reatach_stack(*args, **kwargs):
            if self.flap.get_reveal_progress() != 0:
                return
            self.main_window.flap.disconnect(d)
            self.main_window.flap.remove(self.main_window.flap.get_flap())

            self.preview_layout_cache = self.preview_layout
            self.show()

        # none -> paned
        if (self.preview_layout in [PreviewLayout.FULL_WIDTH, PreviewLayout.HALF_WIDTH, PreviewLayout.HALF_HEIGHT] and
            self.preview_layout_cache is None):
            c = None
            set_flap_mode()
        # horizontal paned -> horizontal paned
        elif (self.preview_layout in [PreviewLayout.FULL_WIDTH, PreviewLayout.HALF_WIDTH] and
            self.preview_layout_cache in [PreviewLayout.FULL_WIDTH, PreviewLayout.HALF_WIDTH]):
            c = None
            set_flap_mode()
        # window -> paned
        elif (self.preview_layout_cache == PreviewLayout.WINDOWED and
              self.preview_layout in [PreviewLayout.FULL_WIDTH, PreviewLayout.HALF_WIDTH, PreviewLayout.HALF_HEIGHT]):
              self.hide()
              self.main_window.flap.set_flap(self.preview_stack)
              c = None
              set_flap_mode()
        # whatever -> paned
        elif (self.preview_layout in [PreviewLayout.FULL_WIDTH, PreviewLayout.HALF_WIDTH, PreviewLayout.HALF_HEIGHT]):
            self.hide()
            c = self.main_window.flap.connect("notify::reveal-progress", set_flap_mode)
        # paned -> windowed
        elif self.preview_layout == PreviewLayout.WINDOWED and  self.preview_layout_cache in [PreviewLayout.FULL_WIDTH, PreviewLayout.HALF_WIDTH, PreviewLayout.HALF_HEIGHT]:
            self.hide()
            d = self.main_window.flap.connect("notify::reveal-progress", reatach_stack)
        # none -> windowed
        else:
            self.main_window.flap.remove(self.main_window.flap.get_flap())
            self.preview_layout_cache = self.preview_layout

    def on_window_title_changed(self, *args, **kwargs):
        self.preview_window_title = self.main_window.get_title() + " - " + _("Preview")

    def on_window_closed(self, window, _event):
        self.main_window.lookup_action("preview").change_state(GLib.Variant.new_boolean(False))



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Gonçalo Silva
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

import math
import webbrowser
from enum import auto, IntEnum

import gi

from apostrophe.preview_renderer import PreviewRenderer
from apostrophe.settings import Settings

gi.require_version('WebKit2', '4.0')
from gi.repository import WebKit2, GLib, Gtk, GObject

from apostrophe.preview_converter import PreviewConverter
from apostrophe.preview_web_view import PreviewWebView


class Step(IntEnum):
    CONVERT_HTML = auto()
    LOAD_WEBVIEW = auto()
    RENDER = auto()


class PreviewHandler:
    """Handles showing/hiding the preview, and allows the user to toggle between modes.

    The rendering itself is handled by `PreviewRendered`. This class handles conversion/loading and
    connects it all together (including synchronization, ie. text changes, scroll)."""

    def __init__(self, window, text_view, flap):
        self.text_view = text_view

        self.web_view = None
        self.web_view_pending_html = None

        self.preview_converter = PreviewConverter()
        self.preview_renderer = PreviewRenderer(
            window, text_view, flap)

        window.connect("style-updated", self.reload)

        self.text_changed_handler_id = None

        self.settings = Settings.new()
        self.web_scroll_handler_id = None
        self.text_scroll_handler_id = None

        self.loading = False
        self.shown = False
        self.preview_visible = False

    def show(self):
        self.__show()
        self.preview_renderer.show()

    def __show(self, html=None, step=Step.CONVERT_HTML):

        if step == Step.CONVERT_HTML:
            # First step: convert text to HTML.
            buf = self.text_view.get_buffer()
            self.preview_converter.convert(
                buf.get_text(buf.get_start_iter(), buf.get_end_iter(), False),
                self.__show, Step.LOAD_WEBVIEW)

        elif step == Step.LOAD_WEBVIEW:
            # Second step: load HTML.
            self.loading = True

            if not self.web_view:
                self.web_view = PreviewWebView()
                self.web_view.get_settings().set_allow_universal_access_from_file_urls(True)
                #TODO: enable devtools on Devel profile
                self.web_view.get_settings().set_enable_developer_extras(True)

                # Show preview once the load is finished
                self.web_view.connect("load-changed", self.on_load_changed)

                # All links will be opened in default browser, but local files are opened in apps.
                self.web_view.connect("decide-policy", self.on_click_link)

                self.web_view.connect("context-menu", self.on_right_click)

            if self.web_view.is_loading():
                self.web_view_pending_html = html
            else:
                self.web_view.load_html(html, "file://localhost/")

        elif step == Step.RENDER:
            # Last step: show the preview. This is a one-time step.
            if not self.preview_visible:
                self.preview_visible = True

                self.text_changed_handler_id = \
                    self.text_view.get_buffer().connect("changed", self.__show)

                self.__show()

            GLib.idle_add(self.web_view.set_scroll_scale, self.text_view.scroll_scale)

            if self.settings.get_boolean("sync-scroll"):
                self.web_scroll_handler_id = \
                    self.web_view.connect("scroll-scale-changed", self.on_web_view_scrolled)
                self.text_scroll_handler_id = \
                    self.text_view.connect("scroll-scale-changed", self.on_text_view_scrolled)

    def reload(self, *_widget, reshow=False):
        if self.preview_visible:
            if reshow:
                self.hide()
            self.show()

    def hide(self):
        if self.preview_visible:
            self.preview_visible = False

            #GLib.idle_add(self.text_view.scroll_scale, self.web_view.get_scroll_scale())
            self.text_view.scroll_scale = self.web_view.get_scroll_scale()

            self.preview_renderer.hide()

            if self.text_scroll_handler_id:
                self.text_view.disconnect(self.text_scroll_handler_id)
                self.text_scroll_handler_id = None
            if self.web_scroll_handler_id:
                self.web_view.disconnect(self.web_scroll_handler_id)
                self.web_scroll_handler_id = None

        if self.text_changed_handler_id:
            self.text_view.get_buffer().disconnect(self.text_changed_handler_id)

        if self.loading:
            self.loading = False

            self.web_view.destroy()
            self.web_view = None

    def update_preview_mode(self):
        self.preview_renderer.update_mode(self.web_view)

    def on_load_changed(self, _web_view, event):
        if event == WebKit2.LoadEvent.FINISHED:
            self.loading = False
            if self.web_view_pending_html:
                self.__show(html=self.web_view_pending_html, step=Step.LOAD_WEBVIEW)
                self.web_view_pending_html = None
            else:
                # we only lazyload the webview once
                if not self.shown:
                    self.preview_renderer.load_webview(self.web_view)
                    self.shown = True
                self.__show(step=Step.RENDER)

    def on_text_view_scrolled(self, _text_view, scale):
        if self.preview_visible and not math.isclose(scale,
                                           self.web_view.get_scroll_scale(),
                                           rel_tol=1e-4):
            self.web_view.set_scroll_scale(scale)

    def on_web_view_scrolled(self, _web_view, scale):
        if self.preview_visible and self.text_view.get_mapped() and not math.isclose(
                scale, self.text_view.scroll_scale, rel_tol=1e-1):
            self.text_view.scroll_scale = scale

    @staticmethod
    def on_click_link(web_view, decision, _decision_type):
        if web_view.get_uri().startswith(("http://", "https://", "www.")):
            webbrowser.open(web_view.get_uri())
            decision.ignore()
            return True

    @staticmethod
    def on_right_click(web_view, context_menu, _event, _hit_test):
        # disable some context menu option
        for item in context_menu.get_items():
            if item.get_stock_action() in [WebKit2.ContextMenuAction.RELOAD,
                                           WebKit2.ContextMenuAction.GO_BACK,
                                           WebKit2.ContextMenuAction.GO_FORWARD,
                                           WebKit2.ContextMenuAction.STOP]:
                context_menu.remove(item)



from gettext import gettext as _

from apostrophe.helpers import user_action


class FormatInserter:
    """Manages insertion of formatting.

    Methods can be called directly, as well as be used as signal callbacks."""

    def insert_italic(self, text_view, _data=None):
        """Insert italic or mark a selection as bold"""

        self.__wrap(text_view, "_", _("italic text"))

    def insert_bold(self, text_view, _data=None):
        """Insert bold or mark a selection as bold"""

        self.__wrap(text_view, "**", _("bold text"))

    def insert_strikethrough(self, text_view, _data=None):
        """Insert strikethrough or mark a selection as strikethrough"""

        self.__wrap(text_view, "~~", _("strikethrough text"))

    def insert_horizontal_rule(self, text_view, _data=None):
        """Insert horizontal rule"""

        text_buffer = text_view.get_buffer()
        with user_action(text_buffer):
            text_buffer.insert_at_cursor("\n\n---\n")
        text_view.scroll_mark_onscreen(text_buffer.get_insert())

    def insert_list_item(self, text_view, _data=None):
        """Insert list item or mark a selection as list item"""

        text_buffer = text_view.get_buffer()
        if text_buffer.get_has_selection():
            (start, end) = text_buffer.get_selection_bounds()
            if start.starts_line():
                with user_action(text_buffer):
                    text = text_buffer.get_text(start, end, False)
                    if text.startswith(("- ", "* ", "+ ")):
                        delete_end = start.copy()
                        delete_end.forward_chars(2)
                        text_buffer.delete(start, delete_end)
                    else:
                        text_buffer.insert(start, "- ")
        else:
            helptext = _("Item")
            text_length = len(helptext)

            cursor_mark = text_buffer.get_insert()
            cursor_iter = text_buffer.get_iter_at_mark(cursor_mark)

            start_ext = cursor_iter.copy()
            start_ext.backward_lines(3)
            text = text_buffer.get_text(cursor_iter, start_ext, False)
            lines = text.splitlines()

            with user_action(text_buffer):
                for line in reversed(lines):
                    if line and line.startswith(("- ", "* ", "+ ")):
                        if cursor_iter.starts_line():
                            text_buffer.insert_at_cursor(line[:2] + helptext)
                        else:
                            text_buffer.insert_at_cursor(
                                "\n" + line[:2] + helptext)
                        break
                    else:
                        if not lines[-1] and not lines[-2]:
                            text_buffer.insert_at_cursor("- " + helptext)
                        elif not lines[-1]:
                            if cursor_iter.starts_line():
                                text_buffer.insert_at_cursor("- " + helptext)
                            else:
                                text_buffer.insert_at_cursor("\n- " + helptext)
                        else:
                            text_buffer.insert_at_cursor("\n\n- " + helptext)
                        break

            self.__select_text(text_view, 0, text_length)

    def insert_ordered_list_item(self, _text_view, _data=None):
        # TODO: implement ordered lists
        pass

    def insert_header(self, text_view, _data=None):
        """Insert header or mark a selection as a list header"""

        text_buffer = text_view.get_buffer()
        with user_action(text_buffer):
            if text_buffer.get_has_selection():
                (start, end) = text_buffer.get_selection_bounds()
                text = text_buffer.get_text(start, end, False)
                text_buffer.delete(start, end)
            else:
                text = _("Header")

            text_buffer.insert_at_cursor("#" + " " + text)

        self.__select_text(text_view, 0, len(text))

    @staticmethod
    def __wrap(text_view, wrap, helptext=""):
        """Inserts wrap format to the selected text
        (helper text when nothing selected)"""
        text_buffer = text_view.get_buffer()
        with user_action(text_buffer):
            if text_buffer.get_has_selection():
                # Find current highlighting
                (start, end) = text_buffer.get_selection_bounds()
                moved = False
                if (start.get_offset() >= len(wrap) and
                        end.get_offset() <= text_buffer.get_char_count() - len(wrap)):
                    moved = True
                    ext_start = start.copy()
                    ext_start.backward_chars(len(wrap))
                    ext_end = end.copy()
                    ext_end.forward_chars(len(wrap))
                    text = text_buffer.get_text(ext_start, ext_end, True)
                else:
                    text = text_buffer.get_text(start, end, True)

                if moved and text.startswith(wrap) and text.endswith(wrap):
                    text = text[len(wrap):-len(wrap)]
                    new_text = text
                    text_buffer.delete(ext_start, ext_end)
                    move_back = 0
                else:
                    if moved:
                        text = text[len(wrap):-len(wrap)]
                    new_text = text.lstrip().rstrip()
                    text = text.replace(new_text, wrap + new_text + wrap)

                    text_buffer.delete(start, end)
                    move_back = len(wrap)

                text_buffer.insert_at_cursor(text)
                text_length = len(new_text)

            else:
                text_buffer.insert_at_cursor(wrap + helptext + wrap)
                text_length = len(helptext)
                move_back = len(wrap)

        cursor_mark = text_buffer.get_insert()
        cursor_iter = text_buffer.get_iter_at_mark(cursor_mark)
        cursor_iter.backward_chars(move_back)
        text_buffer.move_mark_by_name('selection_bound', cursor_iter)
        cursor_iter.backward_chars(text_length)
        text_buffer.move_mark_by_name('insert', cursor_iter)

    @staticmethod
    def __select_text(text_view, offset, length):
        """Selects text starting at the current cursor
        minus offset, length characters."""

        text_buffer = text_view.get_buffer()
        cursor_mark = text_buffer.get_insert()
        cursor_iter = text_buffer.get_iter_at_mark(cursor_mark)
        cursor_iter.backward_chars(offset)
        text_buffer.move_mark_by_name('selection_bound', cursor_iter)
        cursor_iter.backward_chars(length)
        text_buffer.move_mark_by_name('insert', cursor_iter)



from gettext import gettext as _
from gettext import ngettext

from gi.repository import GLib, Gio, Gtk

from apostrophe.settings import Settings
from apostrophe.stats_counter import StatsCounter


class StatsHandler:
    """Shows a default statistic on the stats button,
    and allows the user to toggle which one."""

    # Must match the order/index defined in gschema.xml
    CHARACTERS = 0
    WORDS = 1
    SENTENCES = 2
    PARAGRAPHS = 3
    READ_TIME = 4

    def __init__(self, stats_button, text_view):
        super().__init__()

        self.stats_button = stats_button
        self.stats_button.connect("clicked", self.on_stats_button_clicked)
        self.stats_button.connect("destroy", self.on_destroy)

        self.text_view = text_view
        self.text_view.get_buffer().connect("changed", self.on_text_changed)

        self.popover = None

        self.characters = 0
        self.words = 0
        self.sentences = 0
        self.paragraphs = 0
        self.read_time = (0, 0, 0)

        self.settings = Settings.new()

        self.stats_counter = StatsCounter(self.update_stats)

        self.update_default_stat()

    def on_stats_button_clicked(self, _button):
        self.stats_button.set_state_flags(Gtk.StateFlags.CHECKED, False)

        menu = Gio.Menu()
        stats = self.settings.props.settings_schema.get_key(
            "stat-default").get_range()[1]
        for i, stat in enumerate(stats):
            menu_item = Gio.MenuItem.new(self.get_text_for_stat(i), None)
            menu_item.set_action_and_target_value(
                "app.stat_default", GLib.Variant.new_string(stat))
            menu.append_item(menu_item)
        self.popover = Gtk.Popover.new_from_model(self.stats_button, menu)
        self.popover.connect('closed', self.on_popover_closed)
        self.popover.popup()

    def on_popover_closed(self, _popover):
        self.stats_button.unset_state_flags(Gtk.StateFlags.CHECKED)

        self.popover = None
        self.text_view.grab_focus()

    def on_text_changed(self, buf):
        self.stats_counter.count(
            buf.get_text(
                buf.get_start_iter(),
                buf.get_end_iter(),
                False))

    def get_text_for_stat(self, stat):
        if stat == self.CHARACTERS:
            return ngettext("{:n} Character", "{:n} Characters", self.characters).format(self.characters)
        elif stat == self.WORDS:
            return ngettext("{:n} Word", "{:n} Words", self.words).format(self.words)
        elif stat == self.SENTENCES:
            return ngettext("{:n} Sentence", "{:n} Sentences", self.sentences).format(self.sentences)
        elif stat == self.PARAGRAPHS:
            return ngettext("{:n} Paragraph", "{:n} Paragraphs", self.paragraphs).format(self.paragraphs)
        elif stat == self.READ_TIME:
            return _("{:d}:{:02d}:{:02d} Read Time").format(*self.read_time)
        else:
            raise ValueError("Unknown stat {}".format(stat))

    def update_stats(self, stats):
        (characters, words, sentences, paragraphs, read_time) = stats
        self.characters = characters
        self.words = words
        self.sentences = sentences
        self.paragraphs = paragraphs
        self.read_time = read_time
        self.update_default_stat(False)

    def update_default_stat(self, close_popover=True):
        stat = self.settings.get_enum("stat-default")
        text = self.get_text_for_stat(stat)
        self.stats_button.set_label(text)
        if close_popover and self.popover:
            self.popover.popdown()

    def on_destroy(self, _widget):
        self.stats_counter.stop()



import re
import gi
gi.require_version('Gtk', '3.0')

import logging
logger = logging.getLogger('apostrophe')


class FixTable():

    def __init__(self, TextBuffer):
        self.TextBuffer = TextBuffer

    @staticmethod
    def create_seperator(widths, char):
        """
            Generate a line of + and - as sepeartor

            Example:
            >>> create_separarator([2, 4], '-')
            '+----+------+'
        """
        line = []

        for w in widths:
            line.append("+" + char * (w + 2))

        line.append("+")
        return ''.join(line)

    @staticmethod
    def create_line(columns, widths):
        """
            Crea una fila de la tabla separando los campos con un '|'.

            El argumento `columns` es una lista con las celdas que se
            quieren imprimir y el argumento `widths` tiene el ancho
            de cada columna. Si la columna es mas ancha que el texto
            a imprimir se agregan espacios vacíos.

            Example
                >>> create_line(['nombre', 'apellido'], [7, 10])
                '| nombre  | apellido   |'
        """

        line = zip(columns, widths)
        result = []

        for text, width in line:
            spaces = " " * (width - len(text))
            text = "| " + text + spaces + " "
            result.append(text)

        result.append("|")
        return ''.join(result)

    @staticmethod
    def create_table(content):
        """Imprime una tabla en formato restructuredText.

        El argumento `content` tiene que ser una lista
        de celdas.

        Example:

            >>> print create_table([['software', 'vesion'], ['python', '3.2'],
                ['vim', '7.3']])
+----------+--------+
| software | vesion |
+==========+========+
| python   | 3.2    |
+----------+--------+
| vim      | 7.3    |
+----------+--------+
        """

        # obtiene las columnas de toda la tabla.
        columns = zip(*content)
        # calcula el tamano maximo que debe tener cada columna.
        # replace with len()
        widths = [max([len(x) for x in i]) for i in columns]

        result = []

        result.append(FixTable.create_seperator(widths, '-'))
        print(content, widths)
        result.append(FixTable.create_line(content[0], widths))
        result.append(FixTable.create_seperator(widths, '='))

        for line in content[1:]:
            result.append(FixTable.create_line(line, widths))
            result.append(FixTable.create_seperator(widths, '-'))

        return '\n'.join(result)

    @staticmethod
    def are_in_a_table(current_line):
        "Line in a table?"
        return "|" in current_line or "+" in current_line

    @staticmethod
    def are_in_a_paragraph(current_line):
        "Line in a paragraph?"
        return len(current_line.strip()) >= 1

    def get_table_bounds(self, are_in_callback):
        """
            Gets the row number where the table begins and ends.
            are_in_callback argument must be a function
                    indicating whether a particular line belongs or not
                        to the table to be measured (or create).
                        Returns two values ​​as a tuple
        """
        top = 0

        buf = self.TextBuffer
        start_iter = buf.get_start_iter()
        end_iter = buf.get_end_iter()

        text = self.TextBuffer.get_text(
            start_iter, end_iter, False).split('\n')
        logger.debug(text)
        length = len(text)
        bottom = length - 1

        insert_mark = self.TextBuffer.get_insert()
        insert_iter = self.TextBuffer.get_iter_at_mark(insert_mark)
        current_row_index = insert_iter.get_line()

        for a in range(current_row_index, top, -1):
            if not are_in_callback(text[a]):
                top = a + 1
                break

        for b in range(current_row_index, length):
            if not are_in_callback(text[b]):
                bottom = b - 1
                break

        return top, bottom

    @staticmethod
    def remove_spaces(string):
        """Remove unnecessary spaces"""
        return re.sub("\\s\\s+", " ", string)

    @staticmethod
    def create_separators_removing_spaces(string):
        return re.sub("\\s\\s+", "|", string)

    @staticmethod
    def extract_cells_as_list(string):
        "Extrae el texto de una fila de tabla y lo retorna como una lista."
        string = FixTable.remove_spaces(string)
        return [item.strip() for item in string.split('|') if item]

    @staticmethod
    def extract_table(text, top, bottom):
        full_table_text = text[top:bottom]
        # selecciona solamente las lineas que tienen celdas con texto.
        only_text_lines = [x for x in full_table_text if '|' in x]
        # extrae las celdas y descarta los separadores innecesarios.
        return [FixTable.extract_cells_as_list(x) for x in only_text_lines]

    @staticmethod
    def extract_words_as_lists(text, top, bottom):
        "Genera una lista de palabras para crear una lista."

        lines = text[top:bottom + 1]
        return [FixTable.create_separators_removing_spaces(
            line).split('|') for line in lines]

    def fix_table(self):
        """
            Fix Table, so all columns have the same width (again)

            `initial_row` is a int idicationg the current row index
        """

        cursor_mark = self.TextBuffer.get_insert()
        cursor_iter = self.TextBuffer.get_iter_at_mark(cursor_mark)
        cursor_iter.set_line(cursor_iter.get_line())

        end_line = cursor_iter.copy()
        end_line.forward_to_line_end()

        line_text = self.TextBuffer.get_text(cursor_iter, end_line, False)
        if FixTable.are_in_a_table(line_text):

            # obtiene el indice donde comienza y termina la tabla.
            r1, r2 = self.get_table_bounds(FixTable.are_in_a_table)

            logger.debug('asdasd ')

            # extrae de la tabla solo las celdas de texto
            buf = self.TextBuffer
            start_iter = buf.get_start_iter()
            end_iter = buf.get_end_iter()

            text = self.TextBuffer.get_text(
                start_iter, end_iter, False).split('\n')

            table_as_list = FixTable.extract_table(text, r1, r2)
            logger.debug(table_as_list)
            # genera una nueva tabla tipo restructured text y la dibuja en el
            # buffer.
            table_content = FixTable.create_table(table_as_list)
            logger.debug(table_content)
            # Insert table back into Buffer ...
            self.TextBuffer.insert(start_iter, table_content, -1)
        else:
            logger.debug("Not in a table")
            print("Not in table")



# Copyright (C) 2022, Manuel Genovés <manuel.genoves@gmail.com>
#               2019, Wolf Vollprecht <w.vollprecht@gmail.com>
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
# END LICENSE

"""this dialog adjusts values in gsettings
"""
import webbrowser

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Handy, Gio, GObject
import logging
logger = logging.getLogger('apostrophe')

from apostrophe.settings import Settings

class InputFormat(GObject.Object):
    def __init__(self, name, format, help, **kwargs):
        super().__init__(**kwargs)
        self.name: str = name
        self.format: str = format
        self.help: str = help

@Gtk.Template(resource_path='/org/gnome/gitlab/somas/Apostrophe/ui/Preferences.ui')
class ApostrophePreferencesDialog(Handy.PreferencesWindow):

    __gtype_name__ = "ApostrophePreferencesDialog"

    formats = [
        {
            "name": "Pandoc's Markdown",
            "format": "markdown",
            "help": "https://pandoc.org/MANUAL.html#pandocs-markdown"
        },
        {
            "name": "CommonMark",
            "format": "commonmark",
            "help": "https://commonmark.org"
        },
        {
            "name": "GitHub Flavored Markdown",
            "format": "gfm",
            "help": "https://help.github.com/en/categories/writing-on-github"
        },
        {
            "name": "MultiMarkdown",
            "format": "markdown_mmd",
            "help": "https://fletcherpenney.net/multimarkdown"
        },
        {
            "name": "Plain Markdown",
            "format": "markdown_strict",
            "help": "https://daringfireball.net/projects/markdown"
        }
    ]

    autohide_headerbar_switch = Gtk.Template.Child()
    spellcheck_switch = Gtk.Template.Child()
    input_format_comborow = Gtk.Template.Child()
    bigger_text_switch = Gtk.Template.Child()

    settings = Settings.new()

    def __init__(self):
        super().__init__()
        input_formats = Gio.ListStore.new(InputFormat)

        for i, format in enumerate(self.formats):
            input_formats.append(InputFormat(format["name"],
                                             format["format"],
                                             format["help"]))
            if (format["format"] == self.settings.get_string("input-format")):
                current_format = i

        self.input_format_comborow.bind_name_model(input_formats,
                                                   lambda item, *args: item.name, None, None)

        if current_format:
            self.input_format_comborow.set_selected_index(current_format)

        self.settings.bind("autohide-headerbar",
                           self.autohide_headerbar_switch,
                           "active",
                           Gio.SettingsBindFlags.DEFAULT)

        self.settings.bind("spellcheck",
                           self.spellcheck_switch,
                           "active",
                           Gio.SettingsBindFlags.DEFAULT)

        self.settings.bind("bigger-text",
                           self.bigger_text_switch,
                           "active",
                           Gio.SettingsBindFlags.DEFAULT)

    @Gtk.Template.Callback()
    def on_input_format(self, _widget, _index):
        fmt = self.formats[self.input_format_comborow.get_selected_index()]
        self.settings.set_string("input-format", fmt["format"])

    @Gtk.Template.Callback()
    def on_input_format_help(self, _):
        fmt = self.formats[self.input_format_comborow.get_selected_index()]
        webbrowser.open(fmt["help"])
